import argparse
import json
import os
import shutil
import sys

from IPython.utils.tempdir import TemporaryDirectory
from jupyter_client.kernelspec import KernelSpecManager

kernel_json = {
    "argv": [sys.executable, "-m", "jupyter_workflow.ipython", "-f", "{connection_file}"],
    "display_name": "Jupyter Workflow",
    "language": "python",
}


def install_kernel_spec(user=True, prefix=None):
    with TemporaryDirectory() as td:
        os.chmod(td, 0o755)  # Starts off as 700, not user readable
        with open(os.path.join(td, 'kernel.json'), 'w') as f:
            json.dump(kernel_json, f, sort_keys=True)
        kernel_js = os.path.join(os.path.dirname(__file__), 'kernelspec', 'kernel.js')
        shutil.copy(kernel_js, td)

        print('Installing Jupyter kernel spec')
        KernelSpecManager().install_kernel_spec(td, 'jupyter-workflow', user=user, prefix=prefix)


def _is_root():
    try:
        return os.geteuid() == 0
    except AttributeError:
        return False  # assume not an admin on non-Unix platforms


def main(argv=None):
    ap = argparse.ArgumentParser()
    ap.add_argument('--user', action='store_true',
                    help="Install to the per-user kernels registry. Default if not root.")
    ap.add_argument('--sys-prefix', action='store_true',
                    help="Install to sys.prefix (e.g. a virtualenv or conda env)")
    ap.add_argument('--prefix',
                    help="Install to the given prefix. "
                         "Kernelspec will be installed in {PREFIX}/share/jupyter/kernels/")
    args = ap.parse_args(argv)

    if args.sys_prefix:
        args.prefix = sys.prefix
    if not args.prefix and not _is_root():
        args.user = True

    install_kernel_spec(user=args.user, prefix=args.prefix)


if __name__ == '__main__':
    main()



import asyncio
import itertools
from asyncio import Lock, Queue, FIRST_COMPLETED, Task
from typing import MutableSequence, MutableMapping, Optional, Text, List, Any, cast

from streamflow.core import utils
from streamflow.core.workflow import InputCombinator, Token, Step, InputPort, TerminationToken


class JupyterCartesianProductInputCombinator(InputCombinator):

    def __init__(self,
                 name: Text,
                 step: Optional[Step] = None,
                 ports: Optional[MutableMapping[Text, InputPort]] = None):
        super().__init__(name, step, ports)
        self.lock: Lock = Lock()
        self.queue: Queue = Queue()
        self.terminated: List[Text] = []
        self.token_lists: MutableMapping[Text, List[Any]] = {}

    async def _cartesian_multiplier(self):
        input_tasks = []
        for port_name, port in self.ports.items():
            input_tasks.append(asyncio.create_task(port.get(), name=port_name))
        while True:
            finished, unfinished = await asyncio.wait(input_tasks, return_when=FIRST_COMPLETED)
            input_tasks = list(unfinished)
            for task in finished:
                task_name = cast(Task, task).get_name()
                token = task.result()
                # If a TerminationToken is received, the corresponding port terminated its outputs
                if (isinstance(token, TerminationToken) or
                        (isinstance(token, MutableSequence) and utils.check_termination(token))):
                    self.terminated.append(task_name)
                    # When the last port terminates, the entire combinator terminates
                    if len(self.terminated) == len(self.ports):
                        self.queue.put_nowait([TerminationToken(self.name)])
                        return
                else:
                    # Get all combinations of the new element with the others
                    list_of_lists = []
                    for name, token_list in self.token_lists.items():
                        if name == task_name:
                            list_of_lists.append([token])
                        else:
                            list_of_lists.append(token_list)
                    cartesian_product = list(itertools.product(*list_of_lists))
                    # Put all combinations in the queue
                    for element in cartesian_product:
                        self.queue.put_nowait(list(element))
                    # Put the new token in the related list
                    self.token_lists[task_name].append(token)
                    # Create a new task in place of the completed one
                    input_tasks.append(asyncio.create_task(self.ports[task_name].get(), name=task_name))

    async def _initialize(self):
        # Initialize token lists
        for port in self.ports.values():
            self.token_lists[port.name] = []
        # Retrieve initial input tokens
        input_tasks = []
        for port in self.ports.values():
            input_tasks.append(asyncio.create_task(port.get()))
        inputs = {k: v for (k, v) in zip(self.ports.keys(), await asyncio.gather(*input_tasks))}
        # Check for early termination and put a TerminationToken
        if utils.check_termination(list(inputs.values())):
            self.queue.put_nowait([TerminationToken(self.name)])
        # Otherwise put initial inputs in token lists and in queue and start cartesian multiplier
        else:
            for name, token in inputs.items():
                self.token_lists[name].append(token)
            self.queue.put_nowait(list(inputs.values()))
            asyncio.create_task(self._cartesian_multiplier())

    async def get(self) -> MutableSequence[Token]:
        # If lists are empty it means that this is the first call to the get() function
        async with self.lock:
            if not self.token_lists:
                await self._initialize()
        # Otherwise simply wait for new input tokens
        inputs = await self.queue.get()
        return utils.flatten_list(inputs)



import asyncio
import builtins
import os
from typing import Any, Set, Text, Optional, MutableMapping, MutableSequence

import dill
from IPython.core.compilerop import CachingCompiler
from streamflow.core import utils
from streamflow.core.workflow import Job, Token, Port
from streamflow.data import remotepath
from streamflow.workflow.port import DefaultTokenProcessor

from jupyter_workflow.streamflow import executor
from jupyter_workflow.streamflow.command import JupyterCommandOutput


class ControlTokenProcessor(DefaultTokenProcessor):

    async def compute_token(self, job: Job, command_output: JupyterCommandOutput) -> Token:
        return Token(
            name=self.port.name,
            value=None,
            tag=utils.get_tag(job.inputs),
            job=job.name)


class FileTokenProcessor(DefaultTokenProcessor):

    def __init__(self,
                 port: Port,
                 value: Optional[Text] = None,
                 value_from: Optional[Text] = None):
        super().__init__(port)
        self.value: Optional[Text] = value
        self.value_from: Optional[Text] = value_from

    async def collect_output(self, token: Token, output_dir: Text) -> Token:
        if isinstance(token.job, MutableSequence):
            return token.update(await asyncio.gather(*[asyncio.create_task(
                self.collect_output(t, output_dir)
            ) for t in token.value]))
        elif isinstance(token.value, MutableSequence):
            token_list = await asyncio.gather(*[asyncio.create_task(
                self.collect_output(token.update(t), output_dir)
            ) for t in token.value])
            return token.update([t.value for t in token_list])
        context = self.port.step.context
        path_processor = utils.get_path_processor(self.port.step)
        src_path = token.value
        src_job = context.scheduler.get_job(token.job)
        dest_path = os.path.join(output_dir, path_processor.basename(src_path))
        # Transfer file to local destination
        await self.port.step.context.data_manager.transfer_data(
            src=src_path,
            src_job=src_job,
            dst=dest_path,
            dst_job=None,
            writable=True)
        # Update token
        return token.update(dest_path)

    async def compute_token(self, job: Job, command_output: JupyterCommandOutput) -> Token:
        path_processor = utils.get_path_processor(self.port.step)
        if self.value is not None:
            connector = job.step.get_connector() if job is not None else None
            resources = job.get_resources() or [None]
            if job.output_directory and not path_processor.isabs(self.value):
                pattern = path_processor.join(job.output_directory, self.value)
            else:
                pattern = self.value
            token_value = utils.flatten_list(await asyncio.gather(*[asyncio.create_task(
                remotepath.resolve(
                    connector=connector,
                    target=resource,
                    pattern=pattern
                )) for resource in resources]))
            if len(token_value) == 1:
                token_value = token_value[0]
        else:
            token_value = command_output.user_ns.get(self.value_from)
        if job.output_directory:
            if isinstance(token_value, MutableSequence):
                token_value = [path_processor.join(job.output_directory, t) if not path_processor.isabs(t) else t
                               for t in token_value]
            else:
                if not path_processor.isabs(token_value):
                    token_value = path_processor.join(job.output_directory, token_value)
        return Token(
            name=self.port.name,
            value=token_value,
            job=job.name,
            tag=utils.get_tag(job.inputs))

    async def _register_data(self, job: Job, path: Text):
        connector = job.step.get_connector()
        resources = job.get_resources() or [None]
        register_path_tasks = []
        for resource in resources:
            register_path_tasks.append(asyncio.create_task(
                self.port.step.context.data_manager.register_path(connector, resource, path)))
        await asyncio.gather(*register_path_tasks)

    def get_related_resources(self, token: Token) -> Set[Text]:
        if isinstance(token.job, MutableSequence) or isinstance(token.value, MutableSequence):
            resources = set()
            for t in token.value:
                resources.update(self.get_related_resources(
                    t if isinstance(token.job, MutableSequence) else token.update(t)))
        context = self.port.step.context
        src_job = context.scheduler.get_job(token.job)
        resources = set(src_job.get_resources() if src_job else [])
        data_locations = set()
        for resource in resources:
            data_locations.update(context.data_manager.get_data_locations(resource, token.value))
        resources.update({loc.resource for loc in data_locations})
        return resources

    async def update_token(self, job: Job, token: Token) -> Token:
        if isinstance(token.job, MutableSequence):
            return token.update(await asyncio.gather(*[asyncio.create_task(
                self.update_token(job, t)
            ) for t in token.value]))
        elif isinstance(token.value, MutableSequence):
            token_list = await asyncio.gather(*[asyncio.create_task(
                self.update_token(job, token.update(t))
            ) for t in token.value])
            return token.update([t.value for t in token_list])
        src_job = self.get_context().scheduler.get_job(token.job)
        path_processor = utils.get_path_processor(self.port.step)
        token_value = dill.loads(token.value) if isinstance(token.value, bytes) else token.value
        dest_path = path_processor.join(job.input_directory, os.path.basename(token_value))
        await self.port.step.context.data_manager.transfer_data(
            src=token_value,
            src_job=src_job,
            dst=dest_path,
            dst_job=job)
        return token.update(dill.dumps(dest_path))

    async def weight_token(self, job: Job, token_value: Any) -> int:
        if isinstance(token_value, MutableSequence):
            return sum(await asyncio.gather(*[asyncio.create_task(
                self.weight_token(job, v)
            ) for v in token_value]))
        if job is not None and job.get_resources():
            connector = job.step.get_connector()
            for resource in job.get_resources():
                return await remotepath.size(connector, resource, token_value)
            return 0
        else:
            return await remotepath.size(None, None, token_value) if token_value is not None else 0


class NameTokenProcessor(DefaultTokenProcessor):

    def __init__(self,
                 port: Port,
                 compiler: CachingCompiler,
                 serializer: MutableMapping[Text, Any] = None,
                 value: Optional[Text] = None,
                 value_from: Optional[Text] = None):
        super().__init__(port)
        self.compiler: CachingCompiler = compiler
        self.serializer: MutableMapping[Text, Any] = serializer
        self.value: Optional[Text] = value
        self.value_from: Optional[Text] = value_from

    async def collect_output(self, token: Token, output_dir: Text) -> Token:
        if isinstance(token.job, MutableSequence):
            return await super().collect_output(token, output_dir)
        return token.update(executor.postload(
            compiler=self.compiler,
            name=self.port.name,
            value=([dill.loads(v) for v in token.value]
                   if isinstance(token.value, MutableSequence) else dill.loads(token.value)),
            serializer=self.serializer))

    async def compute_token(self, job: Job, command_output: JupyterCommandOutput) -> Token:
        value = executor.predump(
            compiler=self.compiler,
            name=self.port.name,
            value=(self.value if self.value is not None else
                   command_output.user_ns[self.value_from] if self.value_from in command_output.user_ns else
                   builtins.__dict__.get(self.value_from)),
            serializer=self.serializer)
        token_value = ([dill.dumps(v, recurse=True) for v in value] if isinstance(value, MutableSequence) else
                       dill.dumps(value, recurse=True))
        return Token(
            name=self.port.name,
            value=token_value,
            job=job.name,
            tag=utils.get_tag(job.inputs))


class OutputLogTokenProcessor(DefaultTokenProcessor):

    async def compute_token(self, job: Job, command_output: JupyterCommandOutput) -> Token:
        return Token(
            name=self.port.name,
            value=command_output.value,
            job=job.name,
            tag=utils.get_tag(job.inputs))



import ast
import hashlib
import json
import posixpath
from typing import MutableSequence, Text, Any, MutableMapping, Optional, Tuple, List, Set

from IPython.core.compilerop import CachingCompiler
from IPython.utils.text import DollarFormatter
from streamflow.core import utils
from streamflow.core.context import StreamFlowContext
from streamflow.core.deployment import ModelConfig
from streamflow.core.workflow import Workflow, Target, Step, InputCombinator, InputPort
from streamflow.workflow.combinator import DotProductInputCombinator
from streamflow.workflow.port import DefaultOutputPort, DefaultInputPort, ScatterInputPort, GatherOutputPort
from streamflow.workflow.step import BaseStep

from jupyter_workflow.streamflow import executor
from jupyter_workflow.streamflow.combinator import JupyterCartesianProductInputCombinator
from jupyter_workflow.streamflow.command import JupyterCommand, JupyterCommandToken
from jupyter_workflow.streamflow.token_processor import FileTokenProcessor, NameTokenProcessor, ControlTokenProcessor, \
    OutputLogTokenProcessor


def _build_dependencies(workflow: Workflow, in_step: Step) -> None:
    in_names = list(in_step.input_ports.keys())
    return_values = {s_name: list(s.output_ports.keys()) for s_name, s in reversed(list(workflow.steps.items()))}
    for in_name in in_names:
        for out_step, out_names in return_values.items():
            if in_name in out_names:
                input_port = in_step.input_ports[in_name]
                output_port = workflow.steps[out_step].output_ports[in_name]
                input_port.dependee = output_port
                break


def _build_target(model_name: Text, step_target: MutableMapping[Text, Any]) -> Target:
    target_model = step_target['model']
    return Target(
        model=ModelConfig(
            name=model_name,
            connector_type=target_model['type'],
            config=target_model['config'],
            external=target_model.get('external', False)
        ),
        resources=step_target.get('resources', 1),
        service=step_target.get('service')
    )


def _extract_dependencies(cell_name: Text,
                          compiler: CachingCompiler,
                          ast_nodes: List[Tuple[ast.AST, Text]]) -> MutableSequence[Text]:
    visitor = DependenciesRetriever(cell_name, compiler)
    for node, _ in ast_nodes:
        visitor.visit(node)
    return list(visitor.deps)


def _get_combinator_from_scatter(step: Step,
                                 scatter_ports: MutableMapping[Text, InputPort],
                                 scatter: Optional[MutableMapping[Text, Any]] = None) -> InputCombinator:
    scatter_method = scatter.get('method', 'cartesian')
    combinator_name = utils.random_name()
    if scatter_method == 'cartesian':
        combinator = JupyterCartesianProductInputCombinator(name=combinator_name, step=step)
    else:
        combinator = DotProductInputCombinator(name=combinator_name, step=step)
    if scatter:
        for entry in scatter.get('items') or []:
            if isinstance(entry, str):
                combinator.ports[entry] = scatter_ports[entry]
            else:
                inner_combinator = _get_combinator_from_scatter(
                    step=step,
                    scatter_ports=scatter_ports,
                    scatter=entry)
                combinator.ports[inner_combinator.name] = inner_combinator
    return combinator


def _get_input_combinator(step: Step,
                          scatter: Optional[Any] = None) -> InputCombinator:
    scatter_inputs = _get_scatter_inputs(scatter)
    # If there are no scatter ports in this step, create a single DotProduct combinator
    if not [n for n in scatter_inputs]:
        input_combinator = DotProductInputCombinator(name=utils.random_name(), step=step)
        for port in step.input_ports.values():
            input_combinator.ports[port.name] = port
        return input_combinator
    # If there are scatter ports
    else:
        other_ports = dict(step.input_ports)
        cartesian_combinator = JupyterCartesianProductInputCombinator(name=utils.random_name(), step=step)
        # Separate scatter ports from the other ones
        scatter_ports = {}
        for port_name, port in step.input_ports.items():
            if port_name in scatter_inputs:
                scatter_ports[port_name] = port
                del other_ports[port_name]
        # Choose the right combinator for the scatter ports, based on the scatter method property
        scatter_combinator = _get_combinator_from_scatter(
            step=step,
            scatter_ports=scatter_ports,
            scatter=scatter)
        cartesian_combinator.ports[scatter_combinator.name] = scatter_combinator
        # Create a CartesianProduct combinator between the scatter ports and the DotProduct of the others
        if other_ports:
            dotproduct_name = utils.random_name()
            dotproduct_combinator = DotProductInputCombinator(name=dotproduct_name, step=step)
            dotproduct_combinator.ports = other_ports
            cartesian_combinator.ports[dotproduct_name] = dotproduct_combinator
        return cartesian_combinator


def _get_scatter_inputs(scatter_schema: Optional[MutableMapping[Text, Any]]) -> Set:
    scatter_inputs = set()
    if scatter_schema:
        for entry in scatter_schema.get('items') or []:
            if isinstance(entry, str):
                scatter_inputs.add(entry)
            else:
                scatter_inputs.update(_get_scatter_inputs(entry))
    return scatter_inputs


def _substitute_in_combinator(combinator: InputCombinator, port: InputPort):
    if port.name in combinator.ports:
        combinator.ports[port.name] = port
        return combinator
    combinators = filter(lambda c: isinstance(c, InputCombinator), combinator.ports.values())
    for comb in combinators:
        combinator.ports[comb.name] = _substitute_in_combinator(comb, port)
    return combinator


class NamesStack(object):

    def __init__(self):
        self.stack: List[Set] = [set()]

    def add_scope(self):
        self.stack.append(set())

    def add_name(self, name: Text):
        self.stack[-1].add(name)

    def delete_scope(self):
        self.stack.pop()

    def delete_name(self, name: Text):
        self.stack[-1].remove(name)

    def __contains__(self, name: Text) -> bool:
        for scope in self.stack:
            if name in scope:
                return True
        return False


class DependenciesRetriever(ast.NodeVisitor):

    def __init__(self,
                 cell_name: Text,
                 compiler: CachingCompiler):
        super().__init__()
        self.cell_name: Text = cell_name
        self.compiler: CachingCompiler = compiler
        self.deps: Set[Text] = set()
        self.names: NamesStack = NamesStack()

    def _visit_fields(self, fields):
        for value in fields.values():
            if isinstance(value, list):
                for item in value:
                    if isinstance(item, ast.AST):
                        self.visit(item)
            elif isinstance(value, ast.AST):
                self.visit(value)

    def _visit_Comp(self, node):
        # Add local context
        self.names.add_scope()
        # Extract fields
        fields = {f: v for f, v in ast.iter_fields(node)}
        # Visit the generators field
        for item in fields['generators']:
            if isinstance(item, ast.AST):
                self.visit(item)
        del fields['generators']
        # Visit the other fields
        self._visit_fields(fields)
        # Remove local context
        self.names.delete_scope()

    def _visit_FunctionDef(self, node):
        # Extract fields
        fields = {f: v for f, v in ast.iter_fields(node)}
        # Add name to the context
        self.names.add_name(fields['name'])
        del fields['name']
        # Add local context
        self.names.add_scope()
        # Visit arguments
        self.visit(fields['args'])
        del fields['args']
        # Visit other fields
        self._visit_fields(fields)
        # Remove local context
        self.names.delete_scope()

    def visit_alias(self, node):
        # Extract fields
        fields = {f: v for f, v in ast.iter_fields(node)}
        # If alias is defined add alias, otherwise, add name
        self.names.add_name(fields['asname'] or fields['name'])

    def visit_arg(self, node: ast.arg):
        # Extract fields
        fields = {f: v for f, v in ast.iter_fields(node)}
        # Add arg to context
        self.names.add_name(fields['arg'])
        del fields['arg']
        # Fisit other fields
        self._visit_fields(fields)

    def visit_AsyncFunctionDef(self, node):
        self._visit_FunctionDef(node)

    def visit_Call(self, node) -> Any:
        fields = {f: v for f, v in ast.iter_fields(node)}
        # Check if is a call to `run_cell_magic`
        if isinstance(fields['func'], ast.Attribute) and fields['func'].attr == 'run_cell_magic':
            for _, name, _, _ in DollarFormatter().parse(fields['args'][1].value):
                if name is not None:
                    self.deps.add(name)
        # Visit other fields
        self._visit_fields(fields)

    def visit_ClassDef(self, node):
        # Extract fields
        fields = {f: v for f, v in ast.iter_fields(node)}
        # Add name to the context
        self.names.add_name(fields['name'])
        del fields['name']
        # Add local context
        self.names.add_scope()
        # Visit other fields
        self._visit_fields(fields)
        # Remove local context
        self.names.delete_scope()

    def visit_DictComp(self, node):
        self._visit_Comp(node)

    def visit_ExceptHandler(self, node):
        # Extract fields
        fields = {f: v for f, v in ast.iter_fields(node)}
        # Add name to the context
        if fields['name'] is not None:
            self.names.add_name(fields['name'])
        del fields['name']
        # Add local context
        self.names.add_scope()
        # Visit other fields
        self._visit_fields(fields)
        # Remove local context
        self.names.delete_scope()

    def visit_FunctionDef(self, node):
        self._visit_FunctionDef(node)

    def visit_GeneratorExp(self, node):
        self._visit_Comp(node)

    def visit_ListComp(self, node):
        self._visit_Comp(node)

    def visit_Name(self, node):
        if isinstance(node.ctx, ast.Load) and node.id not in self.names:
            # Skip the 'get_ipython' dependency as it is not serialisable
            if node.id != 'get_ipython':
                self.deps.add(node.id)
        elif isinstance(node.ctx, ast.Store):
            self.names.add_name(node.id)
        self.generic_visit(node)

    def visit_SetComp(self, node):
        self._visit_Comp(node)


class JupyterCell(object):

    def __init__(self,
                 name: Text,
                 code: List[Tuple[ast.AST, Text]],
                 compiler: CachingCompiler,
                 metadata: Optional[MutableMapping[Text, Any]] = None):
        self.name: Text = name
        self.code: List[Tuple[ast.AST, Text]] = code
        self.compiler: CachingCompiler = compiler
        self.metadata: Optional[MutableMapping[Text, Any]] = metadata or {}


class JupyterNotebook(object):

    def __init__(self,
                 cells: List[JupyterCell],
                 autoawait: bool = False,
                 metadata: Optional[MutableMapping[Text, Any]] = None):
        self.cells: List[JupyterCell] = cells
        self.autoawait: bool = autoawait
        self.metadata: Optional[MutableMapping[Text, Any]] = metadata or {}


class JupyterNotebookTranslator(object):

    def __init__(self, context: StreamFlowContext):
        self.context: StreamFlowContext = context

    async def optimize_scatter(self,
                               workflow: Workflow) -> None:
        # Collect gather ports
        gather_ports = set()
        for s in workflow.steps.values():
            for p in s.output_ports.values():
                if isinstance(p, GatherOutputPort):
                    gather_ports.add(posixpath.join(s.name, p.name))
        # Remove gather ports that are workflow output ports
        for output_port in workflow.output_ports.values():
            gather_ports.discard(posixpath.join(output_port.step.name, output_port.name))
        # Remove gather ports that do not have only scatter ports in their successors
        for step in workflow.steps.values():
            for port in step.input_ports.values():
                if port.dependee:
                    dependee = posixpath.join(port.dependee.step.name, port.dependee.name)
                    if dependee in gather_ports:
                        if not isinstance(port, ScatterInputPort):
                            gather_ports.discard(dependee)
        # Collect scatter ports
        scatter_ports = set()
        for s in workflow.steps.values():
            for p in s.input_ports.values():
                if isinstance(p, ScatterInputPort):
                    scatter_ports.add(p)
        # If port depends on a survived gather port, perform optimization
        optimized_out_ports = {}
        for port in scatter_ports:
            if port.dependee:
                dependee = posixpath.join(port.dependee.step.name, port.dependee.name)
                if dependee in gather_ports:
                    in_port = DefaultInputPort(name=port.name, step=port.step)
                    in_port.token_processor = port.token_processor
                    port.step.input_ports[port.name] = in_port
                    port.step.input_combinator = _substitute_in_combinator(
                        combinator=port.step.input_combinator,
                        port=in_port)
                    if dependee in optimized_out_ports:
                        in_port.dependee = optimized_out_ports[dependee]
                    else:
                        out_port = DefaultOutputPort(name=port.dependee.name, step=port.dependee.step)
                        out_port.token_processor = port.dependee.token_processor
                        port.dependee.step.output_ports[port.dependee.name] = out_port
                        in_port.dependee = out_port
                        optimized_out_ports[dependee] = out_port

    async def translate_cell(self,
                             cell: JupyterCell,
                             metadata: Optional[MutableMapping[Text, Any]],
                             autoawait: bool = False) -> Step:
        # Build execution target
        target = metadata.get('target')
        if target is not None:
            if isinstance(target['model'], Text):
                target = {**target, **{'model': metadata['models'][target['model']]}}
            model_name = hashlib.md5(json.dumps(
                obj=target['model'],
                sort_keys=True,
                ensure_ascii=True,
                default=str).encode('ascii')).hexdigest()
            target = _build_target(model_name, target)
        # Extract Python interpreter from metadata
        interpreter = metadata.get('interpreter', 'ipython')
        # Create a step structure
        step = BaseStep(
            name=metadata['cell_id'],
            context=self.context,
            target=target)
        step.workdir = metadata['step'].get('workdir')
        # Process cell inputs
        cell_inputs = metadata['step'].get('in', [])
        input_tokens = {}
        scatter_inputs = _get_scatter_inputs(metadata['step'].get('scatter'))
        for element in cell_inputs:
            # If is a string, it refers to the name of a variable
            if isinstance(element, Text):
                element = {
                    'type': 'name',
                    'name': element
                }
            # Otherwise it must be a dictionary
            if isinstance(element, MutableMapping):
                # Create input port
                name = element.get('name') or element.get('valueFrom')
                if name in scatter_inputs:
                    port = ScatterInputPort(name=name, step=step)
                else:
                    port = DefaultInputPort(name=name, step=step)
                # Get serializer if present
                serializer = (metadata['serializers'][element['serializer']]
                              if isinstance(element['serializer'], Text)
                              else element['serializer']) if 'serializer' in element else None
                # Process port type
                element_type = element['type']
                # If type is equal to `file`, it refers to a file path in the local resource
                if element_type == 'file':
                    port.token_processor = FileTokenProcessor(
                        port=port,
                        value=element.get('value'),
                        value_from=element.get('valueFrom'))
                # If type is equal to `name` or `env`, it refers to a variable
                elif element_type in ['name', 'env']:
                    port.token_processor = NameTokenProcessor(
                        port=port,
                        serializer=serializer,
                        compiler=cell.compiler,
                        value=element.get('value'),
                        value_from=element.get('valueFrom', name))
                # If type is equal to `control`, simply add an empty dependency
                elif element_type == 'control':
                    port.token_processor = ControlTokenProcessor(port=port)
                # Add command token
                input_tokens[name] = JupyterCommandToken(
                    name=name,
                    token_type=element_type,
                    serializer=serializer)
                # Register step port
                step.input_ports[port.name] = port
        # Retrieve inputs automatically if necessary
        if metadata['step'].get('autoin', True):
            input_names = _extract_dependencies(cell.name, cell.compiler, cell.code)
            for name in input_names:
                if name not in step.input_ports:
                    if name in scatter_inputs:
                        port = ScatterInputPort(name=name, step=step)
                    else:
                        port = DefaultInputPort(name=name, step=step)
                    port.token_processor = NameTokenProcessor(
                        port=port,
                        compiler=cell.compiler,
                        value_from=name)
                    input_tokens[name] = JupyterCommandToken(name=name, token_type='name')
                    step.input_ports[port.name] = port
        # If outputs are defined for the current cell
        output_tokens = {}
        if 'out' in metadata['step']:
            for element in metadata['step']['out']:
                # If is a string, it refers to the name of a variable
                if isinstance(element, Text):
                    element = {
                        'type': 'name',
                        'name': element
                    }
                # Otherwise it must be a dictionary
                if isinstance(element, MutableMapping):
                    # Create port
                    name = element.get('name') or element.get('valueFrom')
                    if scatter_inputs:
                        output_port = GatherOutputPort(
                            name=name,
                            step=step,
                            merge_strategy=lambda values: sorted(values, key=lambda t: int(posixpath.basename(t.tag))))
                    else:
                        output_port = DefaultOutputPort(name=name, step=step)
                    # Get serializer if present
                    serializer = (metadata['serializers'][element['serializer']]
                                  if isinstance(element['serializer'], Text)
                                  else element['serializer']) if 'serializer' in element else None
                    # Process port type
                    element_type = element['type']
                    # If type is equal to `file`, it refers to a file path in the remote resource
                    if element_type == 'file':
                        output_port.token_processor = FileTokenProcessor(
                            port=output_port,
                            value=element.get('value'),
                            value_from=element.get('valueFrom'))
                    # If type is equal to `name` or `env`, it refers to a variable
                    elif element_type in ['name', 'env']:
                        output_port.token_processor = NameTokenProcessor(
                            port=output_port,
                            compiler=cell.compiler,
                            serializer=serializer,
                            value_from=element.get('valueFrom', name))
                    # If type is equal to `control`, simply add an empty dependency
                    elif element_type == 'control':
                        output_port.token_processor = ControlTokenProcessor(port=output_port)
                    # Add command token
                    output_tokens[name] = JupyterCommandToken(
                        name=name,
                        token_type=element_type,
                        serializer=serializer)
                    # Register step port
                    step.output_ports[name] = output_port
        # Add output log port
        if scatter_inputs:
            output_log_port = GatherOutputPort(
                name=executor.CELL_OUTPUT,
                step=step,
                merge_strategy=lambda values: sorted(values, key=lambda t: int(posixpath.basename(t.tag))))
        else:
            output_log_port = DefaultOutputPort(name=executor.CELL_OUTPUT, step=step)
        output_log_port.token_processor = OutputLogTokenProcessor(port=output_log_port)
        step.output_ports[executor.CELL_OUTPUT] = output_log_port
        # Set input combinator for the step
        step.input_combinator = _get_input_combinator(
            step=step,
            scatter=metadata['step'].get('scatter'))
        # Create the command to be executed remotely
        step.command = JupyterCommand(
            step=step,
            ast_nodes=cell.code,
            compiler=cell.compiler,
            interpreter=interpreter,
            input_tokens=input_tokens,
            output_tokens=output_tokens,
            autoawait=autoawait)
        return step

    async def translate(self, notebook: JupyterNotebook) -> Workflow:
        # Create workflow
        workflow = Workflow()
        # Parse single cells independently to derive workflow steps
        for cell in notebook.cells:
            step = await self.translate_cell(
                cell=cell,
                metadata={**notebook.metadata, **cell.metadata},
                autoawait=notebook.autoawait)
            _build_dependencies(workflow, step)
            workflow.steps[step.name] = step
        # Apply rewrite rules
        await self.optimize_scatter(workflow)
        # Return the final workflow object
        return workflow



import asyncio
import sys
import time

import streamflow
import traitlets
from ipykernel.ipkernel import IPythonKernel
from ipykernel.jsonutil import json_clean
from ipython_genutils.py3compat import PY3, safe_unicode
from jupyter_client.session import extract_header
from tornado import gen

from jupyter_workflow.config.validator import validate
from jupyter_workflow.ipython.shell import StreamFlowInteractiveShell

try:
    # noinspection PyProtectedMember
    from IPython.core.interactiveshell import _asyncio_runner
except ImportError:
    _asyncio_runner = None


class WorkflowIPythonKernel(IPythonKernel):
    # Set StreamFlow shell
    shell_class = traitlets.Type(StreamFlowInteractiveShell)
    # Kernel info fields
    implementation = 'sf-ipython'
    implementation_version = streamflow.version.VERSION
    language_info = {
        'name': 'python',
        'version': sys.version.split()[0],
        'mimetype': 'text/x-python',
        'codemirror_mode': {
            'name': 'ipython',
            'version': sys.version_info[0]
        },
        'pygments_lexer': 'ipython%d' % (3 if PY3 else 2),
        'nbconvert_exporter': 'python',
        'file_extension': '.py'
    }

    msg_types = IPythonKernel.msg_types + ['workflow_request']

    def init_metadata(self, parent):
        # Call parent functionse
        metadata = super().init_metadata(parent)
        # If StreamFlow has been configured for this cell, store its configuration
        workflow_config = (parent['metadata'] if 'workflow' in parent['metadata'] else
                           parent['content'] if 'workflow' in parent['content'] else
                           None)
        if workflow_config is not None:
            try:
                validate({k: v for k, v in workflow_config['workflow'].items() if k != 'cell_id'})
            except BaseException as e:
                self.log.error(str(e))
                return metadata
            metadata['sf_token'] = self.shell.wf_cell_config.set(workflow_config['workflow'])
        # Return metadata
        return metadata

    def finish_metadata(self, parent, metadata, reply_content):
        # Remove StreamFlow configuration from the `sf_cell_config` attribute if present
        if 'sf_token' in metadata:
            self.shell.wf_cell_config.reset(metadata['sf_token'])
            del metadata['sf_token']
        # Call parent function
        return super().finish_metadata(parent, metadata, reply_content)

    @gen.coroutine
    def workflow_request(self, stream, ident, parent):
        try:
            content = parent['content']
            notebook = content['notebook']
        except:
            self.log.error("Got bad msg: ")
            self.log.error("%s", parent)
            return
        metadata = self.init_metadata(parent)
        reply_content = yield gen.maybe_future(self.do_workflow(notebook, ident, parent))
        sys.stdout.flush()
        sys.stderr.flush()
        if self._execute_sleep:
            time.sleep(self._execute_sleep)
        # Send the reply.
        reply_content = json_clean(reply_content)
        metadata = self.finish_metadata(parent, metadata, reply_content)
        reply_msg = self.session.send(stream, 'execute_reply',
                                      reply_content, parent, metadata=metadata,
                                      ident=ident)
        self.log.debug("%s", reply_msg)

    @gen.coroutine
    def do_workflow(self, notebook, ident, parent):
        shell = self.shell
        reply_content = {}
        if (
            _asyncio_runner
            and shell.loop_runner is _asyncio_runner
            and asyncio.get_event_loop().is_running()
        ):
            coro = shell.run_workflow(notebook)
            coro_future = asyncio.ensure_future(coro)

            with self._cancel_on_sigint(coro_future):
                try:
                    res = yield coro_future
                finally:
                    shell.events.trigger('post_execute')
        else:
            coro = shell.run_workflow(notebook)
            if shell.trio_runner:
                runner = shell.trio_runner
            else:
                runner = shell.loop_runner
            res = runner(coro)
        # Send outputs to cell streams
        for cell_name, content in res.result.items():
            self.session.send(
                self.iopub_thread,
                'stream',
                content={'name': cell_name, 'text': content},
                parent=extract_header(parent),
                ident=ident)
        # Send reply message
        if res.error_before_exec is not None:
            err = res.error_before_exec
        else:
            err = res.error_in_exec
        if res.success:
            reply_content['status'] = 'ok'
        else:
            reply_content['status'] = 'error'
            # noinspection PyProtectedMember
            reply_content.update({
                'traceback': shell._last_traceback or [],
                'ename': str(type(err).__name__),
                'evalue': safe_unicode(err),
            })
        reply_content['execution_count'] = shell.execution_count - 1
        reply_content['payload'] = shell.payload_manager.read_payload()
        shell.payload_manager.clear_payload()
        return reply_content



import argparse
import ast
import asyncio
import builtins
import codeop
import inspect
import io
import json
import os
import shutil
import sys
import traceback
from contextlib import contextmanager, redirect_stdout, redirect_stderr
from tempfile import NamedTemporaryFile
from typing import Dict, MutableSequence

if sys.version_info > (3, 8):
    from ast import Module
else:
    from ast import Module as OriginalModule


    def Module(nodelist, type_ignores):
        return OriginalModule(nodelist)

CELL_OUTPUT = '__JF_CELL_OUTPUT__'
CELL_STATUS = '__JF_CELL_STATUS__'
CELL_LOCAL_NS = '__JF_CELL_LOCAL_NS__'

# Import dill or install it
try:
    import dill
except ImportError:
    import subprocess

    subprocess.call(
        [sys.executable, "-m", "pip", "install", "dill"],
        stdout=subprocess.DEVNULL,
        stderr=subprocess.STDOUT
    )
    import dill

# Define args parser
parser = argparse.ArgumentParser()
parser.add_argument('code_file', metavar='JF_CELL_CODE')
parser.add_argument('output_file', metavar='JF_CELL_OUTPUT')
parser.add_argument('--autoawait', action='store_true')
parser.add_argument('--local-ns-file', nargs='?')
parser.add_argument('--postload-input-serializers', nargs='?')
parser.add_argument('--predump-output-serializers', nargs='?')
parser.add_argument('--output-name', action='append')
parser.add_argument('--workdir', nargs='?')


def _deserialize(path, default=None):
    if path is not None:
        with open(path, "rb") as f:
            return dill.load(f)
    else:
        return default


def _load_element(element):
    return [_load_element(el) for el in element] if isinstance(element, MutableSequence) else dill.loads(element)


def _serialize(compiler, namespace, args):
    if args.predump_output_serializers:
        predump_output_serializers = _deserialize(args.predump_output_serializers)
        namespace = {k: predump(
            compiler=compiler,
            name=k,
            value=v,
            serializer=predump_output_serializers.get(k)) for k, v in namespace.items() if k in args.output_name}
    else:
        namespace = {k: v for k, v in namespace.items() if k in args.output_name}
    with NamedTemporaryFile(delete=False) as f:
        dill.dump(namespace, f, recurse=True)
        return f.name


def compare(code_obj):
    is_async = (inspect.CO_COROUTINE & code_obj.co_flags == inspect.CO_COROUTINE)
    return is_async


def postload(compiler, name, value, serializer):
    if serializer is not None and 'postload' in serializer:
        serialization_context = prepare_ns({name: value})
        postload_module = compiler.ast_parse(
            source=serializer['postload'],
            filename='{name}.postload'.format(name=name))
        for node in postload_module.body:
            mod = Module([node], [])
            code_obj = compiler(mod, '', 'exec')
            exec(code_obj, {}, serialization_context)
        return serialization_context[name]
    else:
        return value


def predump(compiler, name, value, serializer):
    if serializer is not None and 'predump' in serializer:
        serialization_context = prepare_ns({name: value})
        predump_module = compiler.ast_parse(
            source=serializer['predump'],
            filename='{name}.predump'.format(name=name))
        for node in predump_module.body:
            mod = Module([node], [])
            code_obj = compiler(mod, '', 'exec')
            exec(code_obj, {}, serialization_context)
        return serialization_context[name]
    else:
        return value


class RemoteCompiler(codeop.Compile):

    def ast_parse(self, source, filename='<unknown>', symbol='exec'):
        return compile(source, filename, symbol, self.flags | ast.PyCF_ONLY_AST, 1)

    @contextmanager
    def extra_flags(self, flags):
        turn_on_bits = ~self.flags & flags
        self.flags = self.flags | flags
        try:
            yield
        finally:
            self.flags &= ~turn_on_bits


def prepare_ns(namespace: Dict) -> Dict:
    namespace.setdefault('__name__', '__main__')
    namespace.setdefault('__builtin__', builtins)
    namespace.setdefault('__builtins__', builtins)
    if 'get_ipython' in locals():
        namespace.setdefault('get_ipython', locals()['get_ipython'])
    elif 'get_ipython' in globals():
        namespace.setdefault('get_ipython', globals()['get_ipython'])
    return namespace


async def run_code(args):
    output = {CELL_LOCAL_NS: {}}
    command_output = io.StringIO()
    with redirect_stdout(command_output), redirect_stderr(command_output):
        try:
            # Instantiate compiler
            compiler = RemoteCompiler()
            # Deserialize elements
            ast_nodes = _deserialize(args.code_file)
            user_ns = prepare_ns({k: _load_element(v) for k, v in _deserialize(args.local_ns_file, {}).items()})
            # Apply postload serialization hooks if present
            if args.postload_input_serializers:
                postload_input_serializers = _deserialize(args.postload_input_serializers)
                user_ns = {k: postload(
                    compiler=compiler,
                    name=k,
                    value=v,
                    serializer=postload_input_serializers.get(k)
                ) for k, v in user_ns.items()}
            if 'get_ipython' in user_ns:
                user_ns['get_ipython']().user_ns = user_ns
            # Exec cell code
            for node, mode in ast_nodes:
                if mode == 'exec':
                    mod = Module([node], [])
                elif mode == 'single':
                    mod = ast.Interactive([node])
                with compiler.extra_flags(getattr(ast, 'PyCF_ALLOW_TOP_LEVEL_AWAIT', 0x0) if args.autoawait else 0x0):
                    code_obj = compiler(mod, '', mode)
                    asynchronous = compare(code_obj)
                if asynchronous:
                    await eval(code_obj, user_ns, user_ns)
                else:
                    exec(code_obj, user_ns, user_ns)
            # Populate output object
            output[CELL_OUTPUT] = command_output.getvalue().strip()
            if 'get_ipython' in locals():
                output[CELL_OUTPUT] = user_ns['Out'][-1].append(output[CELL_OUTPUT])
            if args.output_name:
                output[CELL_LOCAL_NS] = _serialize(compiler, user_ns, args)
                if args.workdir:
                    dest_path = os.path.join(args.workdir, os.path.basename(output[CELL_LOCAL_NS]))
                    shutil.move(output[CELL_LOCAL_NS], dest_path)
                    output[CELL_LOCAL_NS] = dest_path
            output[CELL_STATUS] = 'COMPLETED'
        except BaseException:
            # Populate output object
            output[CELL_OUTPUT] = command_output.getvalue().strip()
            if output[CELL_OUTPUT]:
                output[CELL_OUTPUT] = output[CELL_OUTPUT] + "\n" + traceback.format_exc().strip()
            else:
                output[CELL_OUTPUT] = traceback.format_exc().strip()
            output[CELL_STATUS] = 'FAILED'
        finally:
            # Save output json to file
            with open(args.output_file, 'w') as f:
                f.write(json.dumps(output))


def main(args):
    # Load arguments
    args = parser.parse_args(args)
    # Run code asynchronously
    if sys.version_info > (3, 7):
        asyncio.run(run_code(args))
    else:
        loop = asyncio.get_event_loop()
        loop.run_until_complete(run_code(args))


if __name__ == "__main__":
    main(sys.argv[1:])



import os
from typing import MutableMapping, Any

from jsonref import loads
from jsonschema import Draft7Validator
from typing_extensions import Text


def load_jsonschema(metadata):
    base_path = os.path.join(
        os.path.dirname(os.path.abspath(__file__)),
        'schemas', metadata['version'])
    filename = os.path.join(base_path, "config_schema.json")
    if not os.path.exists(filename):
        raise Exception(
            'Version in "{}" is unsupported'.format(filename))
    with open(filename, "r") as f:
        return loads(f.read(), base_uri='file://{}/'.format(base_path), jsonschema=True)


def handle_errors(errors):
    errors = list(sorted(errors, key=str))
    if not errors:
        return
    raise Exception(
        "Invalid metadata:\n{error_msg}".format(error_msg=errors[0]))


def validate(workflow_config: MutableMapping[Text, Any]) -> None:
    schema = load_jsonschema(workflow_config)
    validator = Draft7Validator(schema)
    handle_errors(validator.iter_errors(workflow_config))
